/// \file FT245RServer.cpp
///
///  Virtual access to host filesystem via CP/Net-style I/O device.
///
///  \date Apr 22, 2016
///  \author Douglas Miller
///

#include "FT245RServer.h"
#include "host_logging.h"

#include <sys/types.h>
#include <sys/stat.h>
#include <fcntl.h>
#include <unistd.h>
#include <string.h>
#include <poll.h>
#include <sys/ioctl.h>

///
/// User (CP/M) prootocol:
///
/// Init/reset:
///     ; other initialization...
///
/// Send:
///     send "++"
///     send message as pairs of HEX digits...
///     send "--"
///     ; error case
///
/// Receive:
///     wait for "++", discarding illegal characters
///     recv message converting pairs of HEX digits to bytes
///     recv "--"
///     ; error case
///
/// For reference, standard CP/Net message header is:
/// +0  format code (00 = CP/Net send, 01 = response)
/// +1  dest node ID (server or this client, depending on direction)
/// +2  src node ID (this client or server, '')
/// +3  CP/Net, MP/M, CP/M BDOS function number
/// +4  msg size - 1 (00 = 1, FF = 256)
/// +5...   message body
///


FT245RServer::FT245RServer(PropertyUtil::PropertyMapT &props,
				std::string prefix, int cid) :
	ServerDispatch(),
	clientId(cid),
	hostId(0xff),
	header((struct NetworkServer::ndos *)buffer),
	bufIx(-2),
	msgLen(0),
	respLen(0),
	dash(0),
	value(0x0f),
	thread(0),
	_rxfd(-1),
	_txfd(-1)
{
	install_ServerDispatch(props, prefix, cid, hostId);
}

// startup a thread with a pair of pipes.
// Thread handles the protocol, and network/servers.
void FT245RServer::startServer(int rxfd, int txfd) {
	_rxfd = rxfd;
	_txfd = txfd;
	pthread_create(&thread, NULL, protoThread, this);
}

FT245RServer::~FT245RServer() {
	// ensure thread teminates. closing pipes may not be sufficient?
	if (thread != 0) {
		close(_rxfd);
		if (_rxfd != _txfd) {
			close(_txfd);
		}
		pthread_join(thread, NULL);
	}
}

void FT245RServer::resetServer() {
	bufIx = -2;
	msgLen = 0;
	respLen = 0;
	dash = 0;
	value = 0x0f;
}

void *FT245RServer::protoThread(void *v) {
	FT245RServer *thus = (FT245RServer *)v;
	int rc = thus->mainThread(thus->_rxfd, thus->_txfd);
	return (void *)(long)rc;
}

int FT245RServer::mainThread(int rxfd, int txfd) {
	struct pollfd fds[2];
	int nfds = 0;
	int rpfd, spfd;
	fds[nfds].fd = rxfd;
	fds[nfds].events = POLLIN;
	rpfd = nfds++;
	if (rxfd != txfd) {
		fds[nfds].fd = txfd;
		fds[nfds].events = 0;	// might be POLLOUT, if we detrmine it is blocked...
		spfd = nfds++;
	}
	while (1) {
		fds[0].revents = 0;
		fds[1].revents = 0;
		int n = poll(fds, nfds, 0);
		if (n < 0) {
			perror("poll");
			break;
		}
		if (fds[spfd].revents != 0) {
			// probably means HUP...
			//logprintf("HUP spfd\n");
			break;
		}
		if ((fds[rpfd].revents & POLLIN) != 0) {
			uint8_t val = 0;
			fds[rpfd].revents &= ~POLLIN;
			read(rxfd, &val, 1);
			recvByte(val, txfd); // Does send after last byte in message...
		}
		if (fds[rpfd].revents != 0) {
			// probably means HUP...
			//logprintf("HUP rpfd\n");
			break;
		}
	}
	return 0;
}

void FT245RServer::sendResp(int txfd, uint8_t *msgbuf, int len) {
	char hex[4];
	uint8_t *bp = msgbuf;
	int l = len;
	//logprintf("Reply: %02x %02x %02x %02x %02x %02x\n",
	//		msgbuf[0], msgbuf[1], msgbuf[2], msgbuf[3], msgbuf[4], msgbuf[5]);
	write(txfd, "++", 2);
	while (l > 0) {
		sprintf(hex, "%02X", *bp++);
		--l;
		write(txfd, hex, 2);
	}
	write(txfd, "--", 2);
}

void FT245RServer::recvByte(uint8_t val, int txfd) {
	static const char *hex = "0123456789abcdef";
	if (bufIx < 0) {
		if (val == '+') {
			++bufIx;
		} else {
			resetServer();
		}
		return;
	}
	if (val == '+') {
		// possible quick recovery...
		resetServer();
		++bufIx;
		return;
	}
	if (val == '-') {
		// Need exactly two of these...
		if (++dash >= 2) {
			// process message, send result...
			// sanity check against msgLen?

			//logprintf("Recvd: %02x %02x %02x %02x %02x %02x\n",
			//		buffer[0], buffer[1], buffer[2],
			//		buffer[3], buffer[4], buffer[5]);
			// Special case: client wants to know who they are...
			if (header->mfunc == 255) {
				header->mcode |= 1;
				header->mdid = clientId;
				header->msid = hostId;
				header->msize = 0;
				respLen = 1;
			} else {
				respLen = sendMsg(buffer, bufIx);
			}
			if (respLen > 0) {
				sendResp(txfd, buffer, respLen + sizeof(*header));
			}
			resetServer();
		}
		return;
	}
	dash = 0;
	if (!isxdigit(val)) {
		resetServer();
		// TODO: send NAK?
		return;
	}
	// Need a pair of digits...
	value <<= 4;
	value |= strchr(hex, tolower(val)) - hex;
	if (value > 255) {
		buffer[bufIx++] = value & 0xff;
		value = 0x0f;
		if (bufIx == 5) {
			msgLen = header->msize + 1;
		}
	}
}
