/// \file HostFileBdos.cpp
///
///  Virtual access to host filesystem via CP/Net-style I/O device.
///
///  \date Feb 19, 2016
///  \author Douglas Miller
///

#include <sys/types.h>
#include <sys/stat.h>
#include <sys/file.h>
#include <time.h>
#include <fcntl.h>
#include <unistd.h>
#include <dirent.h>
#include <string.h>
#include <fnmatch.h>
#include <stdlib.h>

#include "HostFileBdos.h"
#include "host_logging.h"

struct HostFileBdos::svrcfg HostFileBdos::cfgTab = {0};

void HostFileBdos::initCfg(char tmp, uint8_t sid, uint8_t max) {
	cfgTab.tmp = toupper(tmp) - 'A';
	cfgTab.sts = 0b00010000;	// "ready"
	cfgTab.id = sid;
	cfgTab.max = max;
	cfgTab.cur = 0;
	cfgTab.vec = 0;
	memset(&cfgTab.rid[0], -1, sizeof(cfgTab.rid));
}

// TODO: make thread-safe
int HostFileBdos::addClient(uint8_t cid) {
	int x;
	if (cfgTab.cur >= cfgTab.max) {
		return -1;
	}
	for (x = 0; x < 16; ++x) {
		if ((cfgTab.vec & (1 << x)) == 0) {
			break;
		}
	}
	if (x >= 16) {
		return -1;	// should not happen
	}
	++cfgTab.cur;
	cfgTab.vec |= (1 << x);
	cfgTab.rid[x] = cid;
	return 0;
}

// TODO: make thread-safe
int HostFileBdos::rmClient(uint8_t cid) {
	int x;
	for (x = 0; x < 16; ++x) {
		if (cfgTab.rid[x] == cid) {
			break;
		}
	}
	if (x < 16) {
		--cfgTab.cur;
		cfgTab.vec &= ~(1 << x);
		cfgTab.rid[x] = 0xff;
		return 0;
	}
	return -1;
}

HostFileBdos::HostFileBdos(PropertyUtil::PropertyMapT &props,
			std::vector<std::string> args, uint8_t cltId) :
	NetworkServer(),
	dir(NULL),
	clientId(cltId),
	curDsk(-1),
	curUsr(0),
	curROVec(0),
	curLogVec(0),
	curCompat(0),
	uMask(0)
{
	uMask = umask(0); // always a destructive get...
	umask(uMask);	// restore user's value...
	memset(&curSearch, 0, sizeof(curSearch));
	memset(&curDpb, 0, sizeof(curDpb));
	memset(openFiles, -1, sizeof(openFiles));
	memset(&sFcb, 0, sizeof(sFcb));
	memset(&lsts, -1, sizeof(lsts));
	sFcb.drv= 0x21;
	curDpb.spt = 64; // some number
	curDpb.bsh = DEF_BLS_SH - 7;
	curDpb.blm = (1 << curDpb.bsh) - 1;
	curDpb.dsm = DEF_NBLOCKS - 1;
	curDpb.drm = 63; // some number
	if (curDpb.dsm >= 256) {
		curDpb.exm = (1 << (curDpb.bsh - 4)) - 1;
		phyExt = 8 << curDpb.bsh;
	} else {
		curDpb.exm = (1 << (curDpb.bsh - 3)) - 1;
		phyExt = 16 << curDpb.bsh;
	}
	curDpb.cks = 0; // perhaps should be non-zero, as
			// files can change without notice.
			// But clients shouldn't be handling that.

	// login this client - unless we wait for LOGIN message with PASSWORD...
	// Also, assess thread-safety requirements.
	if (addClient(clientId) < 0) {
		// need to return error and disable this instance...
	}
	char prefix[16];
	sprintf(prefix, "hostfilebdos%02x", cfgTab.id);

	// args[0] is our class name, like argv[0] in main().
	std::string s;
	if (args.size() > 1) {
		s = args[1];
	} else {
		s = prefix;
		s += "_root_dir";
		s = props[s];
	}
	if (s.empty()) {
		s = getenv("HOME");
		s += "/HostFileBdos";
	}
	int rc = mkdir(s.c_str(), 0777);
	if (rc != 0 && errno != EEXIST) {
		logprintf("Cannot create root directory %s\n", s.c_str());
		// what to do... all future accesses will likely return errors.
	}
	logprintf("Creating HostFileBdos device with root dir %s\n", s.c_str());
	dir = strdup(s.c_str());
	// In multi-client environment, temp drive is set only by server init.
	if (cfgTab.max == 1 && args.size() > 2) {
		s = args[2];
		if (s.length() == 2 && isalpha(s[0]) && s[1] == ':') {
			// TODO: thread safety concerns?
			cfgTab.tmp = toupper(s[0]) - 'A';
		}
	}
	std::string pfx = prefix;
	pfx += "_lst";
	int pflen = pfx.length();
	// Could have up to 16 printers assigned...
	PropertyUtil::PropertyMapT::iterator it = props.begin();
	for (; it != props.end(); ++it) {
		if (it->first.compare(0, pflen, pfx) == 0) {
			uint8_t lid;
			unsigned long si = strtoul(it->first.substr(pflen).c_str(), NULL, 16);
			if (si >= 0 && si < 16) {
				lid = si & 0x0f;
			} else {
				logprintf("Invalid List ID %01lx: skipping %s\n", si, it->second.c_str());
				continue;
			}
			if (it->second.at(0) == '!') {
				// Pipe to command...
				s = it->second.substr(1);
				FILE *pp = popen(s.c_str(), "w");
				if (pp == NULL) {
					perror(s.c_str());
				} else {
					lsts[lid] = fileno(pp);
				}
			} else if (it->second.at(0) == '>') {
				// Send to file...
				int flags = O_WRONLY | O_CREAT;
				if (it->second.at(1) == '>') {
					// append...
					flags |= O_APPEND;
					s = it->second.substr(2);
				} else {
					flags |= O_TRUNC;
					s = it->second.substr(1);
				}
				lsts[lid] = open(s.c_str(), flags, 0666);
				if (lsts[lid] < 0) {
					perror(s.c_str());
				}
			} else {
				logprintf("Invalid action: skipping %s\n", it->second.c_str());
				continue;
			}
			logprintf("List %01x: %s\n", lid, it->second.c_str());
		}
	}
}

HostFileBdos::~HostFileBdos() {
	int x;
	for (x = 0; x < DEF_NFILE; ++x) {
		if (openFiles[x].fd >= 0) {
			close(openFiles[x].fd);
			openFiles[x].fd = -1;
		}
	}
	if (curSearch.find.dir) {
		closedir(curSearch.find.dir);
		curSearch.find.dir = NULL;
	}
	if (dir != NULL) {
		free(dir);
		dir = NULL;
	}
	rmClient(clientId);
}

int HostFileBdos::checkRecvMsg(uint8_t clientId, uint8_t *msgbuf, int len) {
	// For HostFileBdos, this indicates an error. The client did not
	// get the response we posted from sendMsg().
	return 0;
}

void dumpMsg(int send, struct NetworkServer::ndos *hdr, uint8_t *msg, int len) {
	if (!send && (hdr->mfunc == 17 || hdr->mfunc == 18)) {
		if (len < 32) {
			logprintf("dirent %d\n", msg[0]);
			return;
		}
		uint8_t *dir = &msg[1];
		--len;
		int x = 0;
		while (len >= 32) {
			if (dir[0] == 0x21) {
				logprintf("dirent %d "
					"%02x "
					"%02x%02x %02x:%02x %02x%02x %02x:%02x %02x "
					"%02x%02x %02x:%02x %02x%02x %02x:%02x %02x "
					"%02x%02x %02x:%02x %02x%02x %02x:%02x %02x\n",
					x,
					dir[0],
					dir[2], dir[1], dir[3], dir[4],
					dir[6], dir[5], dir[7], dir[8], dir[9],
					// dir[10],
					dir[12], dir[11], dir[13], dir[14],
					dir[16], dir[15], dir[17], dir[18], dir[19],
					// dir[20],
					dir[22], dir[21], dir[23], dir[24],
					dir[26], dir[25], dir[27], dir[28], dir[29]
					// dir[30], dir[31]
					);
			} else {
				logprintf("dirent %d "
					"%02x \"%.11s\" %02x %02x %02x %02x "
					"%02x %02x %02x %02x %02x %02x %02x %02x "
					"%02x %02x %02x %02x %02x %02x %02x %02x\n",
					x,
					dir[0], &dir[1],
					dir[12], dir[13], dir[14], dir[15],
					dir[16], dir[17], dir[18], dir[19],
					dir[20], dir[21], dir[22], dir[23],
					dir[24], dir[25], dir[26], dir[27],
					dir[28], dir[29], dir[30], dir[31]);
			}
			dir += 32;
			len -= 32;
			++x;
		}
		return;
	}
	if (hdr->mfunc == 17) { // must be send
		logprintf("MSG %02x: %d: %02x %02x "
			"%02x \"%.11s\" %02x %02x %02x %02x ...\n",
			hdr->mfunc, len, msg[0], msg[1],
			msg[2], &msg[3],
			msg[14], msg[15], msg[16], msg[17]);
		return;
	}
	if (len > 30) { // must have FCB
		logprintf("MSG %02x: %d: %02x "
			"%02x \"%.11s\" %02x %02x %02x %02x "
			"... %02x %02x %02x %02x\n",
			hdr->mfunc, len, msg[0],
			msg[1], &msg[2],
			msg[13], msg[14], msg[15], msg[16],
			msg[33], msg[34], msg[35], msg[36]);
		return;
	}
	logprintf("MSG %02x: %d: %02x %02x [...]\n",
		hdr->mfunc, len, msg[0], msg[1]);
}

int HostFileBdos::sendMsg(uint8_t *msgbuf, int len) {
	struct NetworkServer::ndos *hdr = (struct NetworkServer::ndos *)msgbuf;
	uint8_t *msg = msgbuf + sizeof(*hdr);
	if (bdosFunctions[hdr->mfunc] == NULL) {
		msg[0] = 255;
		msg[1] = 12;
		return 2;
	}
	//dumpMsg(1, hdr, msg, hdr->msize + 1);
	int lr = bdosFunctions[hdr->mfunc](this, msg, len - sizeof(*hdr));
	//dumpMsg(0, hdr, msg, lr);
	return lr;
}

int HostFileBdos::getDPB(uint8_t *msgbuf, int len) {
	memcpy(msgbuf, &curDpb, sizeof(curDpb));
	return sizeof(curDpb);
}
int HostFileBdos::writeProt(uint8_t *msgbuf, int len) {
	curROVec |= (1 << (msgbuf[0] & 0x0f));
	msgbuf[0] = 0;
	return 1;
}
int HostFileBdos::resetDrive(uint8_t *msgbuf, int len) {
	curROVec &= ~(1 << (msgbuf[0] & 0x0f));
	msgbuf[0] = 0;
	return 1;
}
int HostFileBdos::getROVec(uint8_t *msgbuf, int len) {
	msgbuf[0] = (curROVec & 0x0ff);
	msgbuf[1] = ((curROVec >> 8) & 0x0ff);
	return 2;
}
int HostFileBdos::getAllocVec(uint8_t *msgbuf, int len) {
	// TODO: is there anything here?
	memset(msgbuf, 0, 256);
	return 256;
}
int HostFileBdos::getLoginVec(uint8_t *msgbuf, int len) {
	msgbuf[0] = (curLogVec & 0x0ff);
	msgbuf[1] = ((curLogVec >> 8) & 0x0ff);
	return 2;
}

int HostFileBdos::selectDisk(uint8_t *msgbuf, int len) {
	uint8_t d = msgbuf[0];
	msgbuf[0] = 0;
	if (curDsk == d) {
		return 1;
	}
	curDsk = d;
	struct stat stb;
	cpmDrive(fileName, d);
	int rc = stat(fileName, &stb);
	if (rc < 0 && errno == ENOENT) {
		//logprintf("Mkdir %s\n", fileName);
		rc = mkdir(fileName, 0777);
		if (rc == 0) {
			stb.st_mode |= S_IFDIR;
		}
	}
	if (rc < 0 || !S_ISDIR(stb.st_mode)) {
		//logprintf("Seldisk error (%d) %s\n", errno, fileName);
		curLogVec &= ~(1 << d);
		msgbuf[0] = 255;
		return 1;
	}
	//logprintf("Seldisk: %s\n", fileName);
	curLogVec |= (1 << d);
	return 1;
}

int HostFileBdos::openFile(uint8_t *msgbuf, int len) {
	uint8_t u = msgbuf[0] & 0x1f;
	struct fcb *fcb = (struct fcb *)&msgbuf[1];
	msgbuf[0] = 0;
	// ignore password...
	int rc = openFileFcb(fcb, u);
	if (rc < 0) {
		if (rc < -1) {
			msgbuf[0] = 255;
			msgbuf[1] = -rc;
			return 2;
		}
		msgbuf[0] = 255; // a.k.a. File Not Found
		return 1;
	} else {
		return 37;
	}
}

int HostFileBdos::closeFile(uint8_t *msgbuf, int len) {
	int rc;
	//uint8_t u = msgbuf[0] & 0x1f;
	struct fcb *fcb = (struct fcb *)&msgbuf[1];
	bool partial = ATTR_PC(fcb) || (curCompat & COMPAT_PC) != 0;
	clearIntfAttrs(fcb);
	int fd = getFileFcb(fcb);
	if (fd < 0) {
		msgbuf[0] = 255;
		return 1;
	}
	msgbuf[0] = 0;
	if (!fcb->s1[1]) {
		// special truncate for SUBMIT (CCP)
		off_t len = fcb->rc;
		len *= 128;
		rc = ftruncate(fd, len);
		if (rc < 0) {
			logprintf("SUBMIT ftruncate %ld\n", len);
		}
	}
	if (partial) {
		//logprintf("Partial close\n");
		rc = fsync(fd);
		// TODO: will this file ever be closed?
		// could avoid this if we knew no locks were acquired...
		struct flock flk;
		flk.l_type = F_UNLCK;
		flk.l_whence = SEEK_SET;
		flk.l_start = 0;
		flk.l_len = 0; // means "all bytes"
		int rc = fcntl(fd, F_SETLK, &flk);
		// TODO: what errors occur when no locks?
		if (rc < 0) {
			logprintf("fcntl F_UNLCK all err: %d\n", errno);
		}
		rc = flock(fd, LOCK_UN);
		if (rc < 0) {
			logprintf("flock LOCK_UN err: %d\n", errno);
		}
	} else {
		rc = closeFileFcb(fcb);
	}
	if (rc < 0) {
		msgbuf[0] = 255;
		return 1;
	} else {
		return 37;
	}
}

int HostFileBdos::searchNext(uint8_t *msgbuf, int len) {
	//uint8_t u = msgbuf[1] & 0x1f;
	if (curSearch.full) {
		int dc = curSearch.iter & 0x03;
		if (curSearch.iter < curSearch.lastit) {
			++curSearch.iter;
			msgbuf[0] = dc;
			return 1;
		}
		if (curSearch.iter < curSearch.endit) {
			msgbuf[0] = 255;
			return 1;
		}
		// force iter to next record
		curSearch.iter = (curSearch.iter + 0x03) & ~0x03;
	}
	const char *name = doSearch(&curSearch);
	if (!name) {
		msgbuf[0] = 255;
		return 1;
	}
	if (curSearch.full) {
		// curSearch.iter is 0, must be. We know at least one exists.
		msgbuf[0] = fullSearch(&msgbuf[1], &curSearch, name); // fill the dma buf, if possible.
		return 129;
	} else {
		msgbuf[0] = copyOutSearch(&msgbuf[1], name);
		return 33;
	}
}

// CP/M 3 DIR uses fcb->drv == '?' to do full listing,
// and older CP/M STAT uses fcb->ext == '?'.
// However, fcb->drv == '?' is documented in older CP/M
// so we cannot assume CP/M 3 based on that.
int HostFileBdos::searchFirst(uint8_t *msgbuf, int len) {
	//uint8_t d = msgbuf[0] & 0x0f; // should be curDsk...
	uint8_t u = msgbuf[1] & 0x1f;
	struct fcb *fcb = (struct fcb *)&msgbuf[2];
	if ((fcb->drv & 0x7f) == '?') {
		// This means return every dir entry...
		//logprintf("Search with drv = '?'\n");
		memset(&fcb->name[0], '?', 12);
	}
	const char *f = startSearch(fcb, &curSearch, u);
	curLogVec |= (1 << curSearch.drv);
	if (f == NULL) {
		msgbuf[0] = 255;
		return 1;
	}
	if (curSearch.full) {
		// curSearch.iter is 0, must be. We know at least one exists.
		msgbuf[0] = fullSearch(&msgbuf[1], &curSearch, f); // fill the dma buf, if possible.
		return 129;
	} else {
		msgbuf[0] = copyOutSearch(&msgbuf[1], f);
		return 33;
	}
}

// TODO: if this is an open file, then close it?
int HostFileBdos::deleteFile(uint8_t *msgbuf, int len) {
	struct search era;
	uint8_t u = msgbuf[0] & 0x1f;
	msgbuf[0] = 0;
	struct fcb *fcb = (struct fcb *)&msgbuf[1];
	bool delx = ATTR_DELX(fcb);
	clearIntfAttrs(fcb);
	if (fcb->drv == '?') {
		msgbuf[0] = 255;
		msgbuf[1] = 9;
		return 2;
	}
	uint8_t d = getDrv(fcb);
	curLogVec |= (1 << d);
	if ((curROVec & (1 << d)) != 0) {
		msgbuf[0] = 255;
		msgbuf[1] = 2;
		return 2;
	}
	if (delx) {
		// no XFCBs in this implementation,
		// do not remove timestamps.
		return 1;
	}
	fcb->ext = 0;	// in case it was '?'
	// We are supposed to remove NOTHING if ANY matching file is R/O...
	memset(&era, 0, sizeof(era));
	errno = 0;
	const char *name = startSearch(fcb, &era, u);
	int rc = 0;
	while (name) {
		if ((era.mode & (S_IWUSR|S_IWGRP|S_IWOTH)) == 0) {
			rc = -1;
			break;
		}
		++rc;
		name = doSearch(&era);
	}
	endSearch(&era);
	if (rc < 0) {
		msgbuf[0] = 255;
		msgbuf[1] = 3;
		return 2;
	}
	if (rc == 0) {
		msgbuf[0] = 255;
		return 1;
	}
	// Start search over, remove files this time...
	errno = 0;
	name = startSearch(fcb, &era, u);
	rc = 0;
	while (name) {
		++rc;
		unlink(cpmPathFound(&era.find));
		name = doSearch(&era);
	}
	endSearch(&era);
	if (rc == 0) {
		// Probably can't happen since we already checked before.
		msgbuf[0] = 255;
		return 1;
	}
	return 1;
}

int HostFileBdos::readSeq(uint8_t *msgbuf, int len) {
	//uint8_t u = msgbuf[0] & 0x1f;
	struct fcb *fcb = (struct fcb *)&msgbuf[1];
	msgbuf[0] = 0;
	int fd = getFileFcb(fcb);
	if (fd < 0) {
		msgbuf[0] = 9;
		return 1;
	}
	if ((fcb->ext == 0 && fcb->cr == 0) ||
			fcb->cr != fcb->s1[0]) {
		off_t len = fcb->cr;
		len *= 128;
		lseek(fd, len, SEEK_SET);
	}
	int rc = read(fd, &msgbuf[37], 128);
	if (fcb->cr > 127) {
		fcb->cr = 0;
		++fcb->ext;
		fcb->ext &= 0x1f;
	}
	if (rc < 0) {
		msgbuf[0] = 255;
		return 1;
	}
	if (rc == 0) {
		msgbuf[0] = 1;
		return 1;
	}
	++fcb->cr;
	fcb->s1[0] = fcb->cr;
	// fill any partial "sector" with Ctrl-Z, in case it's text.
	memset(&msgbuf[37 + rc], 0x1a, 128 - rc);
	// detect media change?
	return 165;
}

int HostFileBdos::writeSeq(uint8_t *msgbuf, int len) {
	//uint8_t u = msgbuf[0] & 0x1f;
	struct fcb *fcb = (struct fcb *)&msgbuf[1];
	msgbuf[0] = 0;
	int fd = getFileWrFcb(fcb);
	if (fd < 0) {
		if (fd < -1) {
			msgbuf[0] = 255;
			msgbuf[0] = -fd;
			return 2;
		} else {
			msgbuf[0] = 9;
			return 1;
		}
	}
	int rc = write(fd, &msgbuf[37], 128);
	if (fcb->cr > 127) {
		fcb->cr = 0;
		fcb->rc = 0;
		++fcb->ext;
		fcb->ext &= 0x1f;
	}
	++fcb->cr;
	if (fcb->rc < fcb->cr) {
		fcb->rc = fcb->cr;
	}
	if (rc < 0) {
		msgbuf[0] = 255;
		return 1;
	}
	if (rc == 0) {
		msgbuf[0] = 1;
		return 1;
	}
	// fcb->s1[1] = 0; // don't need this here...
	// detect media change?
	return 37;
}

// File is open on successful return.
int HostFileBdos::createFile(uint8_t *msgbuf, int len) {
	uint8_t u = msgbuf[0] & 0x1f;
	struct fcb *fcb = (struct fcb *)&msgbuf[1];
	msgbuf[0] = 0;
	int rc = makeFileFcb(fcb, u);
	if (rc < 0) {
		if (rc < -1) {
			msgbuf[0] = 255;
			msgbuf[1] = -rc;
			return 2;
		}
		msgbuf[0] = 255;
		return 1;
	}
	return 37;
}

int HostFileBdos::renameFile(uint8_t *msgbuf, int len) {
	char newn[sizeof(pathName)];
	uint8_t d;
	uint8_t u = msgbuf[0] & 0x1f;
	struct fcb *fcb = (struct fcb *)&msgbuf[1];
	msgbuf[0] = 0;
	getFileName(fileName, fcb);
	d = getDrv(fcb);
	curLogVec |= (1 << d);
	if ((curROVec & (1 << d)) != 0) {
		msgbuf[0] = 255;
		msgbuf[1] = 2;
		return 2;
	}
	cpmPath(pathName, d, u, fileName);
	fcb = (struct fcb *)&msgbuf[17];
	getFileName(fileName, fcb);
	cpmPath(newn, d, u, fileName);
	int rc = rename(pathName, newn);
	if (rc < 0) {
		// todo: decode errno...
		msgbuf[0] = 255;
		return 1;
	}
	return 1;
}

int HostFileBdos::readRand(uint8_t *msgbuf, int len) {
	//uint8_t u = msgbuf[0] & 0x1f;
	struct fcb *fcb = (struct fcb *)&msgbuf[1];
	msgbuf[0] = 0;
	int fd = getFileFcb(fcb);
	if (fd < 0) {
		msgbuf[0] = 9;
		return 1;
	}
	seekFile(fcb);
	int rc = read(fd, &msgbuf[37], 128);
	seekFile(fcb);
	if (rc < 0) {
		msgbuf[0] = 255;
		return 1;
	}
	if (rc == 0) {
		msgbuf[0] = 1;
		return 1;
	}
	memset(&msgbuf[37 + rc], 0x1a, 128 - rc);
	// detect media change?
	return 37 + 128;
}

int HostFileBdos::writeRand(uint8_t *msgbuf, int len) {
	//uint8_t u = msgbuf[0] & 0x1f;
	struct fcb *fcb = (struct fcb *)&msgbuf[1];
	msgbuf[0] = 0;
	int fd = getFileWrFcb(fcb);
	if (fd < 0) {
		if (fd < -1) {
			msgbuf[0] = 255;
			msgbuf[0] = -fd;
			return 2;
		} else {
			msgbuf[0] = 9;
			return 1;
		}
	}
	seekFile(fcb);
	int rc = write(fd, &msgbuf[37], 128);
	seekFile(fcb);
	if (rc < 0) {
		msgbuf[0] = 255;
		return 1;
	}
	if (rc == 0) {
		msgbuf[0] = 1;
		return 1;
	}
	// detect media change?
	return 37;
}

int HostFileBdos::setRandRec(uint8_t *msgbuf, int len) {
	//uint8_t u = msgbuf[0] & 0x1f;
	struct fcb *fcb = (struct fcb *)&msgbuf[1];
	msgbuf[0] = 0;
	int fd = getFileFcb(fcb);
	if (fd < 0) {
		msgbuf[0] = 9;
		return 1;
	}
	off_t r = lseek(fd, (off_t)0, SEEK_CUR);
	if (r > 0x03ffff * 128) {
		r = 0x03ffff;
	} else {
		r = (r + 127) / 128;
	}
	fcb->rr[0] = r & 0x0ff;
	r >>= 8;
	fcb->rr[1] = r & 0x0ff;
	r >>= 8;
	fcb->rr[2] = r & 0x0ff;
	return 37;
}

int HostFileBdos::compFileSize(uint8_t *msgbuf, int len) {
	struct stat stb;
	uint8_t u = msgbuf[0] & 0x1f;
	struct fcb *fcb = (struct fcb *)&msgbuf[1];
	msgbuf[0] = 0;
	int rc = getFileStat(fcb, u, &stb);
	if (rc < 0) {
		msgbuf[0] = 255;
		return 1;
	}
	off_t r = stb.st_size;
	if (r > 0x03ffff * 128) {
		r = 0x03ffff;
	} else {
		r = (r + 127) / 128;
	}
	fcb->rr[0] = r & 0x0ff;
	r >>= 8;
	fcb->rr[1] = r & 0x0ff;
	r >>= 8;
	fcb->rr[2] = r & 0x003;
	return 37;
}

int HostFileBdos::setFileAttrs(uint8_t *msgbuf, int len) {
	struct stat stb;
	uint8_t u = msgbuf[0] & 0x1f;
	struct fcb *fcb = (struct fcb *)&msgbuf[1];
	int bc = fcb->cr & 0x7f; // 0 means 128
	if (bc == 0) {
		bc = 128;
	}
	// file must not be open?
	int fd = getFileFcb(fcb);
	if (fd >= 0) {
		msgbuf[0] = 0xff;
		msgbuf[1] = fd < -1 ? -fd : 0;
		return 2;
	}
	int rc = getFileStat(fcb, u, &stb);
	if (rc < 0) {
		msgbuf[0] = 255;
		return 1;
	}
	// pathName should now be set, since file was not open
	if (ATTR_BC(fcb)) {
		// Set byte count, essentially truncate last 128-byte record.
		off_t r = stb.st_size;
		if (r > bc) {
			r = ((r + 127) & ~0x7fL) - 128 + bc;
			rc = truncate(pathName, r);
			if (rc < 0) {
				msgbuf[0] = 0xff;
				msgbuf[1] = 1; // probably a R/O file, but not an option.
				return 2;
			}
		}
	}
	bool chg = false;
	bool sys = ((stb.st_mode & S_ISVTX) != 0);
	bool nsys = ATTR_SYS(fcb);
	if (nsys != sys) {
		chg = true;
		if (nsys) {
			stb.st_mode |= S_ISVTX;
		} else {
			stb.st_mode &= ~S_ISVTX;
		}
	}
	bool ro = ((stb.st_mode & (S_IWUSR|S_IWGRP|S_IWOTH)) == 0);
	bool nro = ATTR_RO(fcb);
	if (nro != ro) {
		chg = true;
		if (nro) {
			stb.st_mode &= ~(S_IWUSR|S_IWGRP|S_IWOTH);
		} else {
			stb.st_mode |= (S_IWUSR|S_IWGRP|S_IWOTH) & ~uMask;
		}
	}
	if (chg) {
		rc = chmod(pathName, stb.st_mode);
		if (rc < 0) {
			msgbuf[0] = 0xff;
			msgbuf[1] = 1; // probably a R/O file, but not an option.
			return 2;
		}
	}
	msgbuf[0] = 0;
	return 1;
}
int HostFileBdos::accessDrive(uint8_t *msgbuf, int len) {
	uint16_t vec = msgbuf[0] | (msgbuf[1] << 8);
	curLogVec |= vec;
	msgbuf[0] = 0;
	return 1;
}
int HostFileBdos::freeDrive(uint8_t *msgbuf, int len) {
	uint16_t vec = msgbuf[0] | (msgbuf[1] << 8);
	//logprintf("FREE DRIVE(s) %04x %04x\n", vec, curLogVec);
	// vec==curLogVec: warm boot... via CP/Net 1.2
	closeAll(vec);
	curLogVec &= ~vec;
	curROVec &= ~vec;
	msgbuf[0] = 0;
	return 1;
}
int HostFileBdos::writeRandZF(uint8_t *msgbuf, int len) {
	// Linux zero-fills for us...
	return writeRand(msgbuf, len);
}
int HostFileBdos::lockRec(uint8_t *msgbuf, int len) {
	return recLocking(msgbuf, len, F_WRLCK);
}
int HostFileBdos::unlockRec(uint8_t *msgbuf, int len) {
	return recLocking(msgbuf, len, F_UNLCK);
}
int HostFileBdos::getFreeSp(uint8_t *msgbuf, int len) {
	int cap = ((curDpb.dsm + 1) << curDpb.bsh);
	//logprintf("DISK FREE SPACE %06x\n", cap);
	msgbuf[0] = cap & 0xff;
	msgbuf[1] = (cap >> 8) & 0xff;
	msgbuf[2] = (cap >> 16) & 0xff;
	return 3;
}
int HostFileBdos::flushBuf(uint8_t *msgbuf, int len) {
	int x;
	for (x = 0; x < DEF_NFILE; ++x) {
		if (openFiles[x].fd >= 0) {
			fsync(openFiles[x].fd);
		}
	}
	msgbuf[0] = 0;
	return 1;
}
int HostFileBdos::freeBlks(uint8_t *msgbuf, int len) {
	// CP/Net 3 Warm Boot indicator
	closeAll(0xffff);
	msgbuf[0] = 0;
	return 1;
}
int HostFileBdos::truncFile(uint8_t *msgbuf, int len) {
	int rc;
	uint8_t u = msgbuf[0] & 0x1f;
	struct fcb *fcb = (struct fcb *)&msgbuf[1];
	msgbuf[0] = 0;
	int fd = getFileWrFcb(fcb);
	if (fd >= 0) {
		msgbuf[0] = 0xff;
		msgbuf[1] = fd < -1 ? -fd : 0;
		return 2;
	}
	uint8_t d = getDrv(fcb);
	getFileName(fileName, fcb);
	cpmPath(pathName, d, u, fileName);
	if (access(pathName, W_OK) != 0) {
		msgbuf[0] = 0xff;
		msgbuf[1] = errno == ENOENT ? 0x00 : 0x03;
		return 2;
	}
	struct stat stb;
	stat(pathName, &stb);
	off_t r = fcb->rr[0] | (fcb->rr[1] << 8) | ((fcb->rr[2] & 0x03) << 16);
	if (r > stb.st_size) {
		msgbuf[0] = 0xff;
		msgbuf[1] = 0x00;
		return 2;
	}
	rc = truncate(pathName, r);
	if (rc < 0) {
		msgbuf[0] = 0xff;
		msgbuf[1] = 0xff;
		return 2;
	}
	// TODO: need to update ext,rc,cr? file is not open... as far as we know.
	return 37;
}
int HostFileBdos::login(uint8_t *msgbuf, int len) {
	msgbuf[0] = 0;
	return 1;
}
int HostFileBdos::logoff(uint8_t *msgbuf, int len) {
	msgbuf[0] = 0;
	return 1;
}
int HostFileBdos::setCompAttrs(uint8_t *msgbuf, int len) {
	curCompat = msgbuf[0] & (COMPAT_PRO | COMPAT_PC); // only honor a couple bits...
	msgbuf[0] = 0;
	return 1;
}
int HostFileBdos::getServCfg(uint8_t *msgbuf, int len) {
	memcpy(msgbuf, &cfgTab, sizeof(cfgTab));
	return sizeof(cfgTab);
}
int HostFileBdos::setDefPwd(uint8_t *msgbuf, int len) {
	msgbuf[0] = 0;
	return 1;
}
int HostFileBdos::getDirLab(uint8_t *msgbuf, int len) {
	// CP/M 3 DIR.COM uses full-search, not this, to
	// detect timestamps.
	msgbuf[0] = dirMode;
	return 1;
}
// Get timestamp for given file
int HostFileBdos::readFStamps(uint8_t *msgbuf, int len) {
	uint8_t u = msgbuf[0] & 0x1f;
	struct fcb *fcb = (struct fcb *)&msgbuf[1];
	msgbuf[0] = 0;
	uint8_t d = getDrv(fcb);
	getFileName(fileName, fcb);
	cpmPath(pathName, d, u, fileName);
	if (access(pathName, F_OK) != 0) {
		msgbuf[0] = 0xff;
		msgbuf[1] = 0x00;
		return 2;
	}
	struct stat stb;
	stat(pathName, &stb);
	fcb->ext = 0; // no passwords
	// order is important, overwrites seconds field (not present in CP/M 3 timestamps).
	unix2cpmdate(stb.st_atime, (struct cpmdate *)&fcb->d0[8]);
	unix2cpmdate(stb.st_mtime, (struct cpmdate *)&fcb->d0[12]);
	fcb->cr = 0;
	return 37;
}
int HostFileBdos::getTime(uint8_t *msgbuf, int len) {
	time_t now = time(NULL);
	struct cpmdate *cpm = (struct cpmdate *)&msgbuf[0];
	unix2cpmdate(now, cpm);
	//logprintf("getTime: %04x %02x %02x %02x\n",
	//	msgbuf[0] | (msgbuf[1] << 8), msgbuf[2], msgbuf[3], msgbuf[4]);
	return 5;
}

int HostFileBdos::listOut(uint8_t *msgbuf, int len) {
	uint8_t lid = msgbuf[0];
	if (lid < 16 && lsts[lid] >= 0) {
		// hopefully this doesn't sleep...
		write(lsts[lid], &msgbuf[1], len - 1);
	}
	msgbuf[0] = 0;
	return 1;
}

// *INDENT-OFF*
// '0' means not supported (returns error if called)
int(*HostFileBdos::bdosFunctions[256])(HostFileBdos *, uint8_t *, int) = {
	0, 0, 0, 0, 0, listOut, 0, 0, 0, 0, 0, 0, 0, 0,		// (0-13)
	selectDisk, openFile, closeFile, searchFirst, searchNext, // 14-18
	deleteFile, readSeq, writeSeq, createFile, renameFile,	// 19-23
	getLoginVec,						// 24
	0, 0,							// (25-26)
	getAllocVec, writeProt, getROVec, setFileAttrs, getDPB,	// 27-31
	0,							// (32)
	readRand, writeRand, compFileSize, setRandRec, resetDrive, // 33-37
	accessDrive, freeDrive, writeRandZF,			// 38-40
	0,							// (41)
	lockRec, unlockRec,					// 42-43
	0, 0,							// (44-45)
	getFreeSp,						// 46
	0,							// (47)
	flushBuf,						// 48
	0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0,		// (49-63)
	login, logoff,	// 64, 65: not implemented but must not return error.
	0, 0, 0, 0,						// (66-69)
	setCompAttrs, getServCfg,				// 70, 71
	0, 0, 0, 0, 0, 0, 0, 0,					// (72-79)
	0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0,		// (80-95)
	0, 0,							// (96-97)
	freeBlks, truncFile,					// 98-99
	0, getDirLab, readFStamps, 0, 0,			// (100-104)
	getTime,	// 105: local convenience
	setDefPwd,	// 106: not really implemented...
	0
};
// *INDENT-ON*

int HostFileBdos::cpmDrive(char *buf, int drive) {
	drive &= 0x0f;
	return sprintf(buf, "%s/%c", dir, drive + 'a');
}

int HostFileBdos::cpmFilename(char *buf, int user, char *file) {
	int x = 0;
	user &= 0x1f;
	if (user) {
		x += sprintf(buf + x, "%d:", user);
	}
	x += sprintf(buf + x, "%s", file);
	return x;
}

int HostFileBdos::cpmPath(char *buf, int drive, int user, char *file) {
	int x = cpmDrive(buf, drive);
	buf[x++] = '/';
	x += cpmFilename(buf + x, user, file);
	return x;
}

char *HostFileBdos::cpmNameFound(struct search::find *find) {
	return find->path + find->dirlen + 1;
}

char *HostFileBdos::cpmPathFound(struct search::find *find) {
	return find->path;
}

// Not used, as init is automatic on Search First.
void HostFileBdos::cpmFindInit(struct search::find *find, int drive, char *pattern) {
	if (find->dir) {
		closedir(find->dir);
	}
	strncpy(find->pat, pattern, sizeof(find->pat));
	find->pat[sizeof(find->pat) - 1] = '\0';
	find->dirlen = cpmDrive(find->path, drive);
	find->dir = opendir(find->path);
	// NULL check done later...
	if (!find->dir) {
		//logprintf("no dir: %s\n", find->path);
	}
}

const char *HostFileBdos::cpmFind(struct search::find *find, uint8_t usr) {
	struct dirent *de;
	if (!find->dir) {
		errno = ENXIO;
		return NULL;
	}
	do {
		de = readdir(find->dir);
		if (!de) {
			errno = ENOENT;
			return NULL;
		}
		if (de->d_name[0] == '.') {
			continue;
		}
		//logprintf("looking at %s... %s\n", de->d_name, find->pat);
		// need to prevent "*.*" at user 0 from matching all users!
		if (usr == 0 && (de->d_name[1] == ':' || de->d_name[2] == ':')) {
			continue;
		}
		if (fnmatch(find->pat, de->d_name, 0) == 0) {
			break;
		}
	} while (1);
	sprintf(find->path + find->dirlen, "/%s", de->d_name);
	return cpmNameFound(find);
}

void HostFileBdos::getFileName(char *dst, struct fcb *fcb) {
	int x;
	for (x = 0; x < 8 && (fcb->name[x] & 0x7f) != ' '; ++x) {
		*dst++ = tolower(fcb->name[x] & 0x7f);
	}
	for (x = 8; x < 11 && (fcb->name[x] & 0x7f) != ' '; ++x) {
		if (x == 8) {
			*dst++ = '.';
		}
		*dst++ = tolower(fcb->name[x] & 0x7f);
	}
	*dst++ = '\0';
}

void HostFileBdos::getAmbFileName(char *dst, struct fcb *fcb, uint8_t usr) {
	int x;
	char buf[16];
	char *s = buf;
	int sawQ = 0;
	for (x = 0; x < 8 && (fcb->name[x] & 0x7f) != ' '; ++x) {
		if ((fcb->name[x] & 0x7f) == '?') {
			*s++ = '*';
			sawQ = 1;
			break;
		}
		*s++ = tolower(fcb->name[x] & 0x7f);
	}
	for (x = 8; x < 11 && (fcb->name[x] & 0x7f) != ' '; ++x) {
		if (x == 8) {
			if ((fcb->name[x] & 0x7f) == '?') {
				// special case, need "*.*" to be just "*"
				if (sawQ) {
					break;
				}
			}
			*s++ = '.';
		}
		if ((fcb->name[x] & 0x7f) == '?') {
			*s++ = '*';
			break;
		}
		*s++ = tolower(fcb->name[x] & 0x7f);
	}
	*s++ = '\0';
	cpmFilename(dst, usr, buf);
}

void HostFileBdos::copyOutDir(uint8_t *dma, const char *name) {
	struct fcb *fcb = (struct fcb *)dma;
	const char *t;
	t = strchr(name, ':');
	uint8_t u;
	if (t) {
		u = atoi(name);
		++t;
	} else {
		u = 0;
		t = name;
	}
	fcb->drv = u;	// user code , not drive
	int x = 0;
	while (*t && *t != '.' && x < 8) {
		fcb->name[x++] = toupper(*t++);
	}
	while (x < 8) {
		fcb->name[x++] = ' ';
	}
	while (*t && *t != '.') {
		++t;
	}
	if (*t == '.') {
		++t;
	}
	while (*t && *t != '.' && x < 11) {
		fcb->name[x++] = toupper(*t++);
	}
	while (x < 11) {
		fcb->name[x++] = ' ';
	}
	fcb->ext = 0;
	fcb->s1[0] = 0;	// byte count... overwritten later.
	fcb->s1[1] = 0;
	fcb->rc = 0;
	memset(fcb->d0, 0, sizeof(fcb->d0));
}

int HostFileBdos::copyOutSearch(uint8_t *buf, const char *name) {
	if (*name == ' ') { // DIR LABEL
		struct dirl *dirLab = (struct dirl *)buf;
		sprintf((char *)&dirLab->name[0], "SERVER%02x%c  ",
							cfgTab.id, curSearch.drv + 'A');
		dirLab->ext = dirMode;
		dirLab->drv = 0x20;
		// nothing meaningful for timestamps?
		return curSearch.iter++ & 0x03; // should be "0" (iter == 0)
	} else if (*name == '!') { // SFCB (already populated)
		// this should never happen...
		logprintf("Copy Out SFCB in wrong place\n");
		memcpy(buf, &sFcb, 32);
		return curSearch.iter++ & 0x03; // should be "3"
	}
	copyOutDir(buf, name);
	struct fcb *fcb = (struct fcb *)buf;
	// curSearch.size is only valid if curSearch.ext == '?'.
	if (curSearch.ext == '?') {
		// create fake info in FCB to fool caller into
		// thinking the file is of about the right size.
		// This does not account for sparse files.
		off_t len = curSearch.size;	// len in bytes
		if (len < phyExt * 128) {
			curSearch.size = 0;
		} else {
			curSearch.size -= phyExt * 128;
		}
		len = (len + 127) / 128;	// len in records
		// STAT only cares about (ext & exm), so actually ext can wrap.
		// But, (ext & exm) must reflect the number of
		// logical extents in this dirent.
		int nb = 16, x;
		if (len <= phyExt) {
			nb = (len + curDpb.blm) >> curDpb.bsh;
			fcb->ext = len / 128;	// num logical extents
			fcb->rc = len & 0x7f;
			if (len > 0 && fcb->rc == 0) {
				fcb->rc = 128;
			}
		} else {
			fcb->ext = curDpb.exm;	// should have advancing num in hi bits
			fcb->rc = 128;
		}
		for (x = 0; x < nb; ++x) {
			// just need a non-zero value...
			fcb->d0[x] = 1;
		}
	}
	fcb->s1[0] = curSearch.bc;
	if ((curSearch.mode & S_ISVTX) != 0) {
		SET_SYS(fcb);
	}
	if ((curSearch.mode & (S_IWUSR|S_IWGRP|S_IWOTH)) == 0) {
		SET_RO(fcb);
	}
	// Generally, no one (STAT) cares if this is always 0.
	// CP/Net will place this in a DMA buffer for us.
	return curSearch.iter++ & 0x03;
}

const char *HostFileBdos::commonSearch(struct search *search) {
	if (search->full && search->cpm3) {
		if (search->iter == 0) { // DIR LABEL
			return " ";
		} else if ((search->iter & 0x03) == 0x03) { // SFCB
			return "!";
		}
	}
	if (search->ext == '?' && search->size > 0) {
		// copyOutSearch() should ensure we don't keep landing here,
		// but caller MUST invoke copyOutSearch() each time.
		return cpmNameFound(&search->find);
	}
	const char *f = cpmFind(&search->find, search->usr);
	if (!f) {
		return NULL;
	}
	// return size of file... by some definition...
	struct stat stb;
	// should never fail since we just got filename from directory...
	stat(cpmPathFound(&search->find), &stb);
	if (search->ext == '?') {
		search->size = stb.st_size;
	}
	search->bc = stb.st_size & 0x7f;
	search->mode = stb.st_mode;
	if (search->full && search->cpm3) {
		// SFCB timestamps are only used if the
		// matching FCB is for extent 0. However,
		// we just populate it anyway, especially since
		// extent info is not computed until copyOut.
		int x = search->iter & 0x03; // 0, 1, 2 only
		// order is imperative, overwrite seconds field
		unix2cpmdate(stb.st_atime,
				(struct cpmdate *)&sFcb.fcbs[x].atim);
		unix2cpmdate(stb.st_mtime,
				(struct cpmdate *)&sFcb.fcbs[x].utim);
		sFcb.fcbs[x].pwmode = 0; // clear seconds overrun
	}
	return f;
}

int HostFileBdos::fullSearch(uint8_t *dirbuf, struct search *search, const char *ff) {
	int ix = 0;	// by definition, we are starting with DIRENT 0.
	int saveIter = search->iter;
	search->endit = search->iter + curSearch.maxdc;
	do {
		(void)copyOutSearch(dirbuf, ff);
		dirbuf += 32;
		++ix;
		ff = doSearch(search);
	} while (ff != NULL && ix < search->maxdc); // never includes SFCB
	search->lastit = search->iter; // if >= saveIter+maxdc then no error
	search->iter = saveIter;
	while (ix < 4) {
		if (ix == search->maxdc) { // can only happen for SFCB
			memcpy(dirbuf, &sFcb, 32);
		} else {
			dirbuf[0] = 0xe5;
		}
		dirbuf += 32;
		++ix;
	}
	return search->iter++ & 0x03;
}

const char *HostFileBdos::doSearch(struct search *search) {
	return commonSearch(search);
}

void HostFileBdos::endSearch(struct search *search) {
	if (search->find.dir) {
		closedir(search->find.dir);
		search->find.dir = NULL;
	}
}

const char *HostFileBdos::startSearch(struct fcb *fcb, struct search *search, uint8_t u) {
	char pat[16];
	uint8_t d = fcb->drv & 0x7f;
	search->iter = 0;
	if (d == '?') {
		search->full = 1;
		search->cpm3 = ((fcb->drv & 0x80) != 0); // CP/M 3 sets hi bit
		search->ext = '?';
		search->maxdc = (search->cpm3 ? 3 : 4);
		d = curDsk;
		pat[0] = '*';
		pat[1] = '\0';
	} else {
		search->full = 0;
		search->cpm3 = 0; // don't care if !full
		search->ext = fcb->ext; // might still be '?' (e.g. CP/M 2 STAT.COM)
		if (!d) {
			d = curDsk;
		} else {
			--d;
		}
		getAmbFileName(pat, fcb, u);
	}
	search->drv = d;
	search->usr = u;
	//logprintf("Searching \"%s\" user %d drive %c\n", pat, u, 'A'+d);
	cpmFindInit(&search->find, search->drv, pat);
	return commonSearch(search);
}

void HostFileBdos::seekFile(struct fcb *fcb) {
	int fd = getFileFcb(fcb);
	if (fd < 0) {
		return;
	}
	off_t r = fcb->rr[0] | (fcb->rr[1] << 8) | ((fcb->rr[2] & 0x03) << 16);
	r *= 128;
	(void)lseek(fd, r, SEEK_SET);
	fcb->s1[0] = fcb->cr;
}

int HostFileBdos::newFileFcb() {
	int x;
	for (x = 0; x < DEF_NFILE; ++x) {
		if (openFiles[x].fd < 0) {
			break;
		}
	}
	if (x >= DEF_NFILE) {
		logprintf("Too many open files\n");
		return -1;
	}
	return x;
}

int HostFileBdos::locFileFcb(struct fcb *fcb) {
	if ((fcb->fd ^ fcb->fd_) != 0xffff) {
		return -1;
	}
	int x = fcb->fd;
	if (x < 0 || x >= DEF_NFILE) {
		fcb->fd = fcb->fd_ = 0;
		return -1;
	}
	return x;
}

void HostFileBdos::putFileFcb(struct fcb *fcb, int ix, int fd, uint8_t drv) {
	openFiles[ix].fd = fd;
	if (fd < 0) {
		openFiles[ix].drv = 0;
		fcb->fd = fcb->fd_ = 0;
	} else {
		openFiles[ix].drv = (1 << drv);
		fcb->fd = ix;
		fcb->fd_ = ~ix;
		// Only put file ID in FCB for OPEN/MAKE, the
		// NDOS decides whether to copy it out.
		fcb->rr[0] = ix & 0x0ff;
		fcb->rr[1] = (ix >> 8) & 0x0ff;
	}
}

void HostFileBdos::closeAll(uint16_t vec) {
	int n = 0;
	int x;
	for (x = 0; x < DEF_NFILE; ++x) {
		if (openFiles[x].fd >= 0 && (openFiles[x].drv & vec) != 0) {
			close(openFiles[x].fd);
			openFiles[x].fd = -1;
			++n;
		}
	}
	if (n > 0) {
		logprintf("Close all files %04x - closed %d\n", vec, n);
	}
}

// Returns Unix file descriptor, not CP/M File ID...
int HostFileBdos::getFileFcb(struct fcb *fcb) {
	int x = locFileFcb(fcb);
	if (x < 0) {
		return -1;
	}
	return openFiles[x].fd;
}
int HostFileBdos::getFileWrFcb(struct fcb *fcb) {
	int x = locFileFcb(fcb);
	if (x < 0) {
		return -1;
	}
	if ((openFiles[x].drv & curROVec) != 0) {
		return -2;
	}
	return openFiles[x].fd;
}

// Returns CP/M File Id, or -1 on error.
int HostFileBdos::closeFileFcb(struct fcb *fcb) {
	int x = locFileFcb(fcb);
	if (x < 0) {
		return -1;
	}
	int fd = openFiles[x].fd;
	putFileFcb(fcb, x, -1, 0);
	int rc = close(fd);
	return rc < 0 ? -1 : x;
}

// Returns CP/M File Id, or -1 on error (<0).
int HostFileBdos::openFileFcb(struct fcb *fcb, uint8_t u) {
	struct stat stb;
	bool usr0 = false;
	uint8_t d;
	bool ro = ATTR_OPRO(fcb);
	bool unlk = ATTR_UNLK(fcb);
	clearIntfAttrs(fcb);
	int x = newFileFcb();
	if (x < 0) {
		return -10;	// Open File Limit Exceeded
	}
	getFileName(fileName, fcb);
	d = getDrv(fcb);
	curLogVec |= (1 << d);	// drive must be valid
	cpmPath(pathName, d, u, fileName);
	//logprintf("Opening %s\n", pathName);
	int flags = O_RDWR;
	if (access(pathName, F_OK) < 0 && u) {
		usr0 = true;
		flags = O_RDONLY;
		cpmPath(pathName, d, 0, fileName);
	}
	if (access(pathName, W_OK) != 0) {
		flags = O_RDONLY;
	}
	if (ro) {
		//logprintf("File R/O requested: %s\n", pathName);
		flags = O_RDONLY;
	}
	// NOTE: for this function, file must exist already.
	int fd = open(pathName, flags);
	if (fd < 0) {
		//logprintf("Fail to open %s (%d)\n", pathName, errno);
		return -1;
	}
	fstat(fd, &stb);
	if (usr0 && (stb.st_mode & S_ISVTX) == 0) {
		close(fd);
		return -1;
	}
	if ((curCompat & COMPAT_PRO) != 0) {
		// disable record locking...
		flags |= O_EXCL;
	} else if (!unlk && flags != O_RDONLY) {
		//logprintf("File LOCKED mode: %s\n", pathName);
		int rc = flock(fd, LOCK_EX | LOCK_NB);
		if (rc != 0) {
			close(fd);
			return -5;
		}
		flags |= O_EXCL;
	}
	putFileFcb(fcb, x, fd, d);
	if (usr0) {
		SET_USR0(fcb);
	}
	if ((stb.st_mode & S_ISVTX) != 0) {
		SET_SYS(fcb);
	}
	if ((stb.st_mode & (S_IWUSR|S_IWGRP|S_IWOTH)) == 0) {
		SET_RO(fcb);
	}
	fcb->ext = 0;
	fcb->oflags = flags;
	off_t len = stb.st_size;
	len = (len + 127) / 128;	// num records
	fcb->rc = len > 127 ? 128 : (len & 0x7f);
	if (fcb->cr == 0xff) {
		fcb->cr = stb.st_size & 0x7f; // byte count
	}
	fcb->s1[0] = 0;
	fcb->s1[1] = 0x80;	// flag if close needs to update...
	return x;	// "File ID"
}

// Returns CP/M File ID, or -1 on error (<0).
int HostFileBdos::makeFileFcb(struct fcb *fcb, uint8_t u) {
	uint8_t d;
	bool unlk = ATTR_UNLK(fcb);
	clearIntfAttrs(fcb);
	int x = newFileFcb();
	if (x < 0) {
		return -10;	// Open File Limit Exceeded
	}
	getFileName(fileName, fcb);
	d = getDrv(fcb);
	curLogVec |= (1 << d);
	if ((curROVec & (1 << d)) != 0) {
		return -2;	// R/O Disk
	}
	cpmPath(pathName, d, u, fileName);
	int flags = O_RDWR;
	int fd = open(pathName, flags | O_CREAT | O_EXCL, 0666);
	if (fd < 0) {
		if (errno == ENOENT) {
			// can't happen?
		}
		return -1;
	}
	if ((curCompat & COMPAT_PRO) != 0) {
		// disable record locking...
		flags |= O_EXCL;
	} else if (!unlk) {
		//logprintf("File LOCKED mode: %s\n", pathName);
		// should never fail...
		int rc = flock(fd, LOCK_EX | LOCK_NB);
		if (rc != 0) {
			close(fd);
			return -5;
		}
		flags |= O_EXCL;
	}
	putFileFcb(fcb, x, fd, d);
	fcb->ext = 0;
	fcb->rc = 0;
	fcb->oflags = flags;
	fcb->s1[1] = 0x80;	// flag to update on close
	return x;	// "File ID"
}

void HostFileBdos::unix2cpmdate(time_t unx, struct cpmdate *cpm) {
	struct tm tmv;
	localtime_r(&unx, &tmv);
	// date is UTC, time is local, need to reconcile them...
	unx += tmv.tm_gmtoff;
	cpm->date = unx / 86400 - 2922 + 1;
	// the rest are BCD fields...
	cpm->hour = ((tmv.tm_hour / 10) << 4) | (tmv.tm_hour % 10);
	cpm->min = ((tmv.tm_min / 10) << 4) | (tmv.tm_min % 10);
	cpm->sec = ((tmv.tm_sec / 10) << 4) | (tmv.tm_sec % 10);
}

int HostFileBdos::recLocking(uint8_t *msgbuf, int len, int mode) {
	//uint8_t u = msgbuf[0] & 0x1f;
	struct fcb *fcb = (struct fcb *)&msgbuf[1];
	uint16_t *fid = (uint16_t *)&msgbuf[37];
	int fd = getFileFcb(fcb);
	if (fd <= 0) {
		msgbuf[0] = 9;
		return 1;
	}
	if (*fid != fcb->fd) {
		msgbuf[0] = 13;
		return 1;
	}
	if ((fcb->oflags & O_EXCL) == 0) {
		struct flock flk;
		flk.l_type = mode;
		flk.l_whence = SEEK_SET;
		flk.l_start = fcb->rr[0] | (fcb->rr[1] << 8) | ((fcb->rr[2] & 0x03) << 16);
		flk.l_len = 128;
		int rc = fcntl(fd, F_SETLK, &flk);
		if (rc < 0) {
			if (errno == EACCES || errno == EAGAIN) {
				msgbuf[0] = 8;
			} else {
				msgbuf[0] = 12;
			}
			return 1;
		}
	}
	msgbuf[0] = 0;
	return 37;
}

void HostFileBdos::clearIntfAttrs(struct fcb *fcb) {
	uint16_t attrs = 0;
	int x;
	for (x = 0; x < 11; ++x) {
		if ((fcb->name[x] & 0x80) != 0) {
			attrs |= (1 << x);
		}
		fcb->name[x] &= 0x7f;
	}
}

int HostFileBdos::getFileStat(struct fcb *fcb, uint8_t u, struct stat *stb) {
	int fd = getFileFcb(fcb);
	int rc;
	if (fd < 0) {
		getFileName(fileName, fcb);
		cpmPath(pathName, getDrv(fcb), u, fileName);
		rc = stat(pathName, stb);
	} else {
		rc = fstat(fd, stb);
	}
	if (rc < 0) {
		return -1;
	}
	return 0;
}
