/// \file ServerDispatch.cpp
///
///  Virtual access to host filesystem via CP/Net-style I/O device.
///
///  \date Feb 19, 2016
///  \author Douglas Miller
///

#include "ServerDispatch.h"
#include "host_logging.h"

///
/// For reference, standard CP/Net message header is:
/// +0  format code (00 = CP/Net send, 01 = response)
/// +1  dest node ID (server or this client, depending on direction)
/// +2  src node ID (this client or server, '')
/// +3  CP/Net, MP/M, CP/M BDOS function number
/// +4  msg size - 1 (00 = 1, FF = 256)
/// +5...   message body
///

ServerDispatch::ServerDispatch()
{
}

ServerDispatch::~ServerDispatch() {
}

void ServerDispatch::addServer(uint8_t serverId, NetworkServer *server) {
	servers[serverId] = server;
}

void ServerDispatch::install_ServerDispatch(PropertyUtil::PropertyMapT &props,
					std::string prefix, uint8_t cid, uint8_t &hid) {
	std::string s;
	PropertyUtil::PropertyMapT::iterator it = props.begin();
	prefix += "server";
	int pflen = prefix.length();
	for (; it != props.end(); ++it) {
		// property syntax: cpnetdevice_server## = ClassId [args...]
		// where '##' is serverId in hex.
		if (it->first.compare(0, pflen, prefix) == 0) {
			uint8_t sid;
			unsigned long si = strtoul(it->first.substr(pflen).c_str(), NULL, 16);
			// TODO: check for conflicts/duplicates?
			if (si >= 0 && si < 255) {
				sid = si & 0x0ff;
			} else {
				logprintf("Invalid Server ID %02lx: skipping %s\n", si, it->second.c_str());
				continue;
			}
			logprintf("Server %02x: %s\n", sid, it->second.c_str());
			std::vector<std::string> args = PropertyUtil::splitArgs(it->second);

			if (args[0].compare("HostFileBdos") == 0) {
				// In this environment, only one client is allowed.
				// TODO: make temp drive configurable...
				HostFileBdos::initCfg('P', sid, 1);
				NetworkServer *nws = new HostFileBdos(props, args, cid);
				addServer(sid, nws);
				if (hid == 0xff) {
					hid = sid;
				}

//			One instance of SocketServer could handle multiple/all servers...
//			} else if args[0].compare("Socket") == 0) {
//				NetworkServer *nws = new SocketServer(props, args, sid, cid);
//				addServer(sid, nws);
			}
		}
	}
}

void ServerDispatch::swapIds(struct NetworkServer::ndos *header) {
	uint8_t temp = header->mdid;

	header->mdid = header->msid;
	header->msid = temp;
}

int ServerDispatch::checkRecvMsg(uint8_t clientId, uint8_t *msgbuf, int len) {
	// TODO: interate over servers and call checkRecvMsg on each...
	// server message should include these already? but clientId may not be
	// known to servers.
	//      header->mcode = 1; // CP/Net response
	//      header->mdid = clientId;
	//      header->msid = serverId;
	//      header->mfunc = 0; // TODO: save last func? get from
	//      header->msize = len - 1;
	return 0;
}

int ServerDispatch::sendMsg(uint8_t *msgbuf, int len) {
	struct NetworkServer::ndos *hdr = (struct NetworkServer::ndos *)msgbuf;
	NetworkServer *nws = servers[hdr->mdid];

	if (nws == NULL) {
		logprintf("Attempting to send to non-existent server %02x\n",
			hdr->mdid);
		return -1;	// reasonable?
	}
	//logprintf("Message: %02x %02x %02x %02x %02x : %02x\n",
	//	msgbuf[0], msgbuf[1], msgbuf[2], msgbuf[3], msgbuf[4], msgbuf[5]);
	int rlen = nws->sendMsg(msgbuf, len);
	if (rlen > 0) {
		// setup response
		swapIds(hdr);
		hdr->mcode |= 0x01;
		hdr->msize = rlen - 1;
	}
	return rlen;
}
